#include "Python.h"


#include "structmember.h"
#include "xlistobject.h"
#include "xlistinfo.h"
#include "xiterobject.h"
#include "xquery.h"
#include "expatll.h"
#include "regexpr.h"
#include "kernel.h"

#ifdef _WIN32
#include <time.h>
#else
#include <sys/time.h>
#endif

static char kernel__doc__[] =
"kernel is the core of Satine XML data binding. It provides the xlist type.\n\
and basic functions to work with xlists.";

/* We link this module statically for convenience.  If compiled as a shared
   library instead, some compilers don't allow addresses of Python objects
   defined in other libraries to be used in static initializers here.  The
   DEFERRED_ADDRESS macro is used to tag the slots where such addresses
   appear; the module init function must fill in the tagged slots at runtime.
   The argument is for documentation -- the macro ignores it.
*/
#define DEFERRED_ADDRESS(ADDR) 0

typedef struct _frame {
    PyObject_VAR_HEAD
    struct _frame *f_back;	/* previous frame, or NULL */
    PyObject *f_code;	/* code segment */
    PyObject *f_builtins;	/* builtin symbol table (PyDictObject) */

} PyFrameObject;

py2xml_strict = 0;

PyObject* kernel_spaces;
PyObject* kernel_uri2mods;
PyObject* kernel_mod2uri;
PyObject* default_prefixes;
PyObject* default_uris;

PyObject* xmlcomment_type;
PyObject* xmldeclaration_type;
PyObject* xmlPI_type;
PyObject* xmlprolog_type;
static int kernel_verbosity = 0;

const int opt_xmlComment = 1;
const int opt_xmlEntity = 2;
const int opt_xmlCData = 4;
const int opt_xmlDeclaration = 8;
const int opt_xmlPI = 16;

const int opt_xmlStrict = 255;
const int opt_autoCreationOff = 256;

static struct {
	PyObject* mod;	
	PyObject* uri;
	PyObject* prefix;
} pending_space;

static PyObject* 
getTag_FromType(PyTypeObject* type) {
	PyObject* tag;
	tag = PyDict_GetItemString(type->tp_dict, "__tag__");
	if (tag == NULL)
		tag = PyString_FromString(type->tp_name);
	else
		if (!PyString_Check(tag)) {
			Satine_WriteInfo("In type %s, __tag___ field %s is not a valid identifier\n", type->tp_name,
				PyString_AsString(PyObject_Repr(tag)));
			PyErr_SetString(PyExc_AttributeError, "__tag__ must be a string");
			return NULL;
		}
	return tag;
}

static PyObject*
kernel_reg_space_c(PyObject *mod, PyObject * uri, PyObject* prefix) {
	PyObject* mod_dict;
	PyObject* mod_name;
	PyObject* key;
	PyObject* tag;
	PyObject* type;
	PyObject* xlist_dict;
	PyObject* uri2mods_list;
	int pos;

	mod_dict = PyModule_GetDict(mod);
	mod_name = PyObject_GetAttrString(mod, "__name__");
	
	PyDict_SetItemString(mod_dict, "__uri__", uri);
	PyDict_SetItemString(mod_dict, "__prefix__", prefix);

	xlist_dict = PyDict_GetItem(kernel_spaces, uri);
	if (!xlist_dict) 
		xlist_dict = PyDict_New();

	Satine_WriteInfo("Satine: import module '%s', uri='%s'\n", PyString_AsString(mod_name), PyString_AsString(uri));
	
	pos = 0;
	while (PyDict_Next(mod_dict, &pos, &key, &type)) {
		if (!PyType_Check(type))
			continue;
		if (!PyType_IsSubtype((PyTypeObject*) type, &PyXList_Type))
			continue;
		
		tag = getTag_FromType((PyTypeObject*) type);
		if (!tag)
			return NULL;

		PyDict_SetItem(xlist_dict, tag, type);
	}

	PyDict_SetItem(kernel_spaces, uri, xlist_dict);
	uri2mods_list = PyDict_GetItem(kernel_uri2mods, uri);
	if (!uri2mods_list) {
		uri2mods_list = PyList_New(0);
		PyDict_SetItem(kernel_uri2mods, uri, uri2mods_list);
	}
	if (!PySequence_Contains(uri2mods_list, mod))
		PyList_Append(uri2mods_list, mod);
	PyDict_SetItem(kernel_mod2uri, mod_name, uri);

	if (PyString_Size(prefix)) {
		PyDict_SetItem(default_uris, prefix, uri);
		PyDict_SetItem(default_prefixes, uri, prefix);
	}
	PyObject_SetAttrString(mod, "space", xlist_dict);

	Py_INCREF(Py_None);
	return Py_None;
}

static PyObject *
kernel_reg_space(PyObject* self, PyObject* args) {
	PyObject* mod;
	PyObject* uri;
	PyObject* prefix;

	if (!PyArg_ParseTuple(args, "O!O!O!", &PyModule_Type, &mod,
			&PyString_Type, &uri, &PyString_Type, &prefix)) 
		return NULL;

	return kernel_reg_space_c(mod, uri, prefix);
}

static void 
uriAndMod_random_names(PyObject** uri, PyObject** modname) {
	long num;
	time_t tm;
	char nm[64];

	time(&tm);
	num = tm*100+rand();
	sprintf(nm, "http://satine.sourceforge.net/temp/dyn/%i", num);

	*uri = PyString_FromString(nm);
	sprintf(nm, "satine.temp.dynmodule%i", num);
	*modname = PyString_FromString(nm);
}

static void 
mod_random_name(PyObject** modname) {
	long num;
	time_t tm;
	char nm[64];

	time(&tm);
	num = tm*100+rand();
	sprintf(nm, "satine.temp.dynmodule%i", num);
	*modname = PyString_FromString(nm);
}

static PyObject *
kernel_new_space_c(PyObject* uri, PyObject* prefix, PyObject* name) {
	PyObject* mod;
	PyObject* xlist_dict;
	PyObject* uri2mods_list;
	PyObject* sys_modules;
	char* name_c;

	name_c = PyString_AsString(name);

	mod = PyModule_New(name_c);
	PyObject_SetAttrString(mod, "__uri__", uri) ;
	PyObject_SetAttrString(mod, "__prefix__", prefix) ;

	xlist_dict = PyDict_GetItem(kernel_spaces, uri);
	if (!xlist_dict) {
		xlist_dict = PyDict_New();
		PyDict_SetItem(kernel_spaces, uri, xlist_dict);
		Py_DECREF(xlist_dict);
	}

	uri2mods_list = PyDict_GetItem(kernel_uri2mods, uri);
	if (!uri2mods_list) {
		uri2mods_list = PyList_New(0);
		PyDict_SetItem(kernel_uri2mods, uri, uri2mods_list);
		Py_DECREF(uri2mods_list);
	}
	if (!PySequence_Contains(uri2mods_list, mod))
		PyList_Append(uri2mods_list, mod);
	PyDict_SetItem(kernel_mod2uri, name, uri);
	PyObject_SetAttrString(mod, "__space__", xlist_dict);
	sys_modules = PySys_GetObject("modules");
	PyDict_SetItem(sys_modules, name, mod);

	Py_DECREF(mod);
	return mod;
}


static PyObject *
kernel_new_space(PyObject *self, PyObject *args) {
	PyObject* uri;
	PyObject* prefix;
	PyObject* name = NULL;

	if (!PyArg_ParseTuple(args, "O!O!|O!", &PyString_Type, &uri, &PyString_Type, &prefix, 
			&PyString_Type, &name)) 
		return NULL;

	return kernel_new_space_c(uri,prefix,name);
}

static PyObject *
kernel_new_type_c(PyObject *uri, PyObject *tag,
				PyObject* __attrs__, PyObject* __items__, PyObject* name) {

	PyObject* classes;
	PyObject* cls;
	PyObject* dict;
	PyObject* mod_name;
	PyObject* mod;
	PyObject* mod_list;
	PyObject* xlist_dict;
	int len;

	if (name == Py_None) {
		Py_INCREF(tag);
		name = tag;
	}

	mod_list = PyDict_GetItem(kernel_uri2mods, uri);
	if (!mod_list)
		return NULL;
	len = PyList_GET_SIZE(mod_list);
	mod = PyList_GET_ITEM(mod_list, 0);
	mod_name = PyObject_GetAttrString(mod, "__name__");

	dict = PyDict_New();
	classes = PyTuple_New(1);
	PyTuple_SetItem(classes, 0, (PyObject*) &PyXList_Type);
	cls = PyClass_New(classes, dict, name);

	PyObject_SetAttrString(cls, "__module__", mod_name);
	PyObject_SetAttr(mod, name, cls);
	
	xlist_dict = PyDict_GetItem(kernel_spaces, uri);
	PyDict_SetItem(xlist_dict, tag, cls);

	Py_DECREF(dict);
	Py_DECREF(classes);
	return cls;
}

static PyObject *
kernel_new_type(PyObject *self, PyObject *args) {
	PyObject* name = NULL;
	PyObject* tag;
	PyObject* mod;
	PyObject* __attrs__ = Py_None;
	PyObject* __items__ = Py_None;
	PyObject* etable = NULL;

	if (!PyArg_ParseTuple(args, "O!O!|O!O!O!", &PyString_Type, &tag, &PyModule_Type, &mod, 
			&PyString_Type, &__attrs__, &PyString_Type, &__items__,	&PyString_Type, &name)) 
		return NULL;
	return kernel_new_type_c(tag,mod,__attrs__,__items__,name);
}

static PyObject *
kernel_get_type_c(PyObject* uri, PyObject* tag) {
	PyObject* space;
	PyObject* type;

	space = PyDict_GetItem(kernel_spaces, uri);
	if (!space) {
		PyErr_SetString(PyExc_AttributeError, "not registed uri");
		return NULL;
	}
	
	type = PyDict_GetItem(space, tag);
	if (!type) {
		PyErr_SetString(PyExc_AttributeError, "undefined tag");
		return NULL;
	}

	Py_INCREF(type);
	return type;
}

static PyObject *
kernel_get_type(PyObject *self, PyObject *args) {
	PyObject* tag;
	PyObject* uri;
	
	if (!PyArg_ParseTuple(args, "OO", &tag, &uri)) 
		return NULL;

	return kernel_get_type_c(tag, uri);	
}

static PyObject *
kernel_get_space(PyObject *self, PyObject *args) {
	PyObject* uri;
	PyObject* space;
	
	if (!PyArg_ParseTuple(args, "O", &uri)) 
		return NULL;

	space = PyDict_GetItem(kernel_spaces, uri);
	if (!space) {
		PyErr_SetString(PyExc_AttributeError, "not registed uri");
		return NULL;
	}
	
	Py_INCREF(space);
	return space;
}

static PyObject*
kernel_is_xlist(PyObject *self, PyObject *xlist) {
	int cmp = PyXList_Check(xlist); 
	PyObject* res = PyInt_FromLong(cmp);
	Py_INCREF(res);
	return res;	
}

static PyObject*
kernel_uri_from_prefix(PyObject* prefix, PyObject* suggested_uris, int options) {
	PyObject* uri = NULL;
	PyObject* mod_list;
	PyObject* modname = NULL;

	if (suggested_uris)
		uri = PyDict_GetItem(suggested_uris, prefix);

	if (!uri)  
		uri = PyDict_GetItem(default_uris, prefix);

	if (!uri) {
		if (options & opt_autoCreationOff) {
			PyErr_SetObject(PyExc_TypeError, 
				PyString_FromFormat(
					"namespace with prefix %s  not found among registered modules",
					PyString_AsString(prefix)));
			return NULL;
		}
		else
			uriAndMod_random_names(&uri, &modname);
	}

	mod_list = PyDict_GetItem(kernel_uri2mods, uri);
	if (!mod_list) {
		if (!modname)
			mod_random_name(&modname);

		kernel_new_space_c(uri, prefix, modname);
//		PyDict_SetItem(default_uris, prefix, uri);
		PyDict_SetItem(default_prefixes, uri, prefix);
		Py_DECREF(modname);
		Py_DECREF(uri);
	}

	if (suggested_uris)
		PyDict_SetItem(suggested_uris, prefix, uri);

	return uri;
}


static int 
kernel_bind_ns(PyObject* attr, PyObject* suggested_uris) {
	char* key_c;
	char* prefix_c;
	PyObject* key;
	PyObject* prefix;
	PyObject* value;

	prefix = PyTuple_GET_ITEM(attr, 0);
	key = PyTuple_GET_ITEM(attr, 1);
	value = PyTuple_GET_ITEM(attr, 2);
	
	key_c = PyString_AsString(key);
	if (!strcmp(key_c, "xmlns")) {
		PyDict_SetItemString(suggested_uris, "", value);
		return 1;
	}
	if (prefix == Py_None)
		return 0;

	prefix_c = PyString_AsString(prefix);

	if (prefix)
		if (!strcmp(prefix_c, "xmlns")) {
			PyDict_SetItem(suggested_uris, key, value);
			return 1;
		}
	return 0;
}

static int 
kernel_is_a_ns(struct XmlItem_c* attr) {
	char* tag;
	char* prefix;
	
	tag = PyString_AsString(attr->tag);
	if (!strcmp(tag, "xmlns")) {
		return 1;
	}
	prefix = PyString_AsString(attr->prefix);
	if (prefix)
		if (!strcmp(prefix, "xmlns")) {
			return 1;
		}
	return 0;
}

static void 
kernel_fill_ns(struct XmlItem_c* item, PyObject* suggested_uris) {
	PyObject* attrs;
	PyObject* attr;
	int i, size;

	if (item->attrs == Py_None)
		return;

	attrs = item->attrs;
	size = PyList_GET_SIZE(attrs);
	for (i=0; i < size; ++i) { 
		attr = PyList_GET_ITEM(attrs, i);
		if (kernel_bind_ns(attr, suggested_uris)) {
			PySequence_DelItem(attrs, i);
			i--;
			size--;
		}
	}
}



static int 
kernel_fill_attrs(PyXListObject* xlist, struct XmlItem_c* item, PyObject* suggested_uris, int options) {
	PyObject* attr;
	PyObject* attrs;
	PyObject* fullname;
	PyObject* prefix;
	PyObject* stdprefix;
	PyObject* key;
	PyObject* value;
	PyObject* uri;
	int res;
	int i;
	int size;

	if (item->attrs == Py_None)
		return 0;

	attrs = item->attrs;
	size = PyList_GET_SIZE(attrs);
	for (i=0; i < size; ++i) {
		attr = PyList_GET_ITEM(attrs, i);

		prefix = PyTuple_GET_ITEM(attr, 0);
		key = PyTuple_GET_ITEM(attr, 1);
		value = PyTuple_GET_ITEM(attr, 2);

		if (prefix != Py_None) {
			uri = kernel_uri_from_prefix(prefix, suggested_uris, options);
			if (uri == NULL)
				return -1;

			stdprefix = PyDict_GetItem(default_prefixes, uri);
			if (stdprefix)
				fullname = Attr_FullName(stdprefix, key);
			else
				fullname = Attr_FullName(prefix, key);
		}
		else {
			Py_INCREF(key);
			fullname = key;
		}

		res = PyDict_SetItem(xlist->attrs, fullname, value);
		Py_DECREF(fullname);
		if (res == -1)
			return -1;
	}
	return 0;
}

static int
kernel_XmlItem2XmlDecl(PyObject* dest, struct XmlItem_c* item, 
							 PyObject* suggested_uris, int options) {

	PyObject* xmldecl;

	xmldecl = PyObject_CallObject(xmldeclaration_type, NULL);

	if (kernel_fill_attrs((PyXListObject*)xmldecl, item, suggested_uris, options) == -1) {
		Py_DECREF(item->attrs);
		return -1;
	}
	
	PyList_Append(dest, xmldecl);
	Py_INCREF(Py_None);
	PyList_Append(dest, Py_None);
	Py_DECREF(item->attrs);
	return 0;

}

static int
kernel_XmlItem2XmlPI(PyObject* dest, struct XmlItem_c* item, 
							 PyObject* suggested_uris, int options) {

	PyObject* xmlPI;

	xmlPI = PyObject_CallObject(xmlPI_type, NULL);

	if (kernel_fill_attrs((PyXListObject*)xmlPI, item, suggested_uris, options) == -1)
		return -1;
	
	PyObject_SetAttrString(xmlPI, "__PI", item->tag);

	PyList_Append(dest, xmlPI);
	Py_INCREF(Py_None);
	PyList_Append(dest, Py_None);

	Py_DECREF(item->prefix);
	Py_DECREF(item->tag);
	Py_DECREF(item->attrs);

	return 0;

}

static int
kernel_XmlItem2PyXListObject(PyObject* dest, struct XmlItem_c* item, 
							 PyObject* suggested_uris, int options) {
	PyObject* uri;
	PyObject* type;
	PyObject* xlist;

	kernel_fill_ns(item, suggested_uris);
	uri = kernel_uri_from_prefix(item->prefix, suggested_uris, options);
	if (uri == NULL) {
		Py_DECREF(item->tag);
		Py_DECREF(item->prefix);
		return -1;
	}

	type = kernel_get_type_c(uri, item->tag); 
	if (!type) {
		if (options & opt_autoCreationOff) {
			PyErr_SetObject(PyExc_TypeError, 
				PyString_FromFormat(
					"xlist with uri %s and tag %s not found among registered modules",
					PyString_AsString(uri),
					PyString_AsString(item->tag)));
			Py_DECREF(item->tag);
			Py_DECREF(item->prefix);
			return -1;
		}
		else
			type = kernel_new_type_c(uri, item->tag, Py_None, Py_None, Py_None);
	}

	Satine_WriteMoreInfo("xlist creation %s\n", ((PyTypeObject*)type)->tp_name);

	xlist = PyObject_CallObject(type, NULL);
	if (xlist == NULL) {
		Py_DECREF(item->tag);
		Py_DECREF(item->prefix);
		Satine_WriteMoreInfo("error during instantiation of %s\n", ((PyTypeObject*)type)->tp_name);
		return -1;
	}

	Satine_WriteMoreInfo("xlist end creation \n");
	if (kernel_fill_attrs((PyXListObject*)xlist, item, suggested_uris, options) == -1) {
		Py_DECREF(item->tag);
		Py_DECREF(item->prefix);
		return -1;
	}

	
	PyList_Append(dest, xlist);
	Py_DECREF(xlist);
	Py_DECREF(item->prefix);
	Py_DECREF(item->tag);
	Py_DECREF(item->attrs);
	Satine_WriteMoreInfo("xlist attr end %s %i\n", PyString_AsString(PyObject_Repr(xlist)), xlist->ob_refcnt);
	return 0;

}


static int
kernel_XmlItem2XmlComment(PyObject* dest, struct XmlItem_c* item, 
							 int options) {
	PyObject* xmlcomment;
	xmlcomment = PyObject_CallObject(xmlcomment_type, NULL);
	PyList_Append(dest, xmlcomment);
	PyList_Append(dest, item->value);
	Py_DECREF(item->value); 
	Py_INCREF(Py_None);
	PyList_Append(dest, Py_None);
	return 0;
}

static void
kernel_put_None(PyObject* dest, struct XmlItem_c* item) {
	PyList_Append(dest, Py_None);
}

static void
kernel_put_Data(PyObject* dest, struct XmlItem_c* item) {
	PyList_Append(dest, item->value);
	Py_DECREF(item->value); 
}

static void 
kernel_setpos_c(PyObject* pos, int offset) {
	PyObject* po = PyInt_FromLong(offset);
	PyList_Append(pos, po);
	Py_DECREF(po);
}

#define optional(code) if (!(options & code)) break

static PyObject*
kernel_parsexml_c(char* str, int len, PyObject* suggested_uri, int options) {
	char* next;
	char* start;
	char* end;
	PyObject* dest = PyXList_New();
	PyObject* pos = PyList_New(0);
	PyObject* res;
	struct XmlItem_c item;
	int error;

	if (suggested_uri == NULL)
		suggested_uri = PyDict_New();
	else
		Py_INCREF(suggested_uri);

	start = str;
	end = start+len;

	while (start < end) {
		error = 0;
		expatll_parse(&item, start, end, &next);

		switch (item.type) {
			case XmlItem_partial: end = start; break;
			case XmlItem_startTag: error = kernel_XmlItem2PyXListObject(dest, &item, suggested_uri, options);
									   kernel_setpos_c(pos, start-str);break;
			case XmlItem_endTag: kernel_put_None(dest, &item); 
								kernel_setpos_c(pos, start-str);break;
			case XmlItem_emptyElement: kernel_XmlItem2PyXListObject(dest, &item, suggested_uri, options); 
								kernel_setpos_c(pos, start-str);
								kernel_put_None(dest, &item); 
								kernel_setpos_c(pos, start-str); break;
			case XmlItem_data: kernel_put_Data(dest, &item); 
								kernel_setpos_c(pos, start-str); break;
			case XmlItem_comment: optional(opt_xmlComment); 
								kernel_XmlItem2XmlComment(dest, &item, options);
								kernel_setpos_c(pos, start-str); 
								kernel_setpos_c(pos, start-str); 
								kernel_setpos_c(pos, start-str); break;
			case XmlItem_XmlDecl: optional(opt_xmlDeclaration); 
								kernel_XmlItem2XmlDecl(dest, &item, suggested_uri, options);
								kernel_setpos_c(pos, start-str); 
								kernel_setpos_c(pos, start-str); break;
			case XmlItem_PI: optional(opt_xmlPI); 
								kernel_XmlItem2XmlPI(dest, &item, suggested_uri, options);
								kernel_setpos_c(pos, start-str); 
								kernel_setpos_c(pos, start-str); break;
		}

		start = next;
	}
	Py_DECREF(suggested_uri);

	if (error == -1) {
		Satine_WriteMoreInfo("error in parsexml %s\n",str);

		Py_DECREF(dest);
		Py_DECREF(pos);
		return NULL;
	}
	
	kernel_setpos_c(pos, start-str); 
	res = PyTuple_New(2);
//	Py_INCREF(dest);
//	Py_INCREF(pos);

	PyTuple_SetItem(res, 0, dest);
	PyTuple_SetItem(res, 1, pos);

	Satine_WriteMoreInfo("parsexml_c end %s\n", PyString_AsString(PyObject_Repr(res)));

	return res;
}

static PyObject*
kernel_parsexml(PyObject* self, PyObject *args, PyObject* kwds) {
	PyObject* suggested_uri = NULL;
	char* s_buffer;
	int s_size;
	int options = 0;
    static char *kwlist[] = {"doc", "uris", "opt", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s#|O!i", kwlist, 
                                     &s_buffer, &s_size, &PyDict_Type, &suggested_uri,
									 &options))
		return NULL;

	Satine_WriteMoreInfo("xlist kernel %s\n", s_buffer);
	

	return kernel_parsexml_c(s_buffer, s_size, suggested_uri, options);
}

PyObject*
kernel_xml2seq_c(char* s_buffer, int s_size, PyObject* suggested_uri, int options);

void 
kernel_cache_data(PyObject* ptr, char** buffer, char** pos, int* buffer_len) {
	char* data;
	int size;
	int i;
	char* new_buffer;
	char c;

	data = PyString_AS_STRING(ptr);
	size = PyString_GET_SIZE(ptr);

	if ((*pos-*buffer)+size > *buffer_len) {
		*buffer_len += ((*pos-*buffer)+size);

		new_buffer = malloc(sizeof(char)*(*buffer_len));
		memcpy(new_buffer, *buffer, *pos-*buffer);

		free(*buffer);
		*pos += new_buffer - *buffer;
		*buffer = new_buffer;
	}
	for (i=0; i < size; ++i) {
		c = data[i];
		if (c > ' ' || c == '\\') {
			**pos = c;
			(*pos)++;
		}
		else if (c == ' ' && (*pos != *buffer) && ((*pos)[-1] != ' ')) {
			**pos = c;
			(*pos)++;
		}
	}

}

PyObject*
kernel_seq2py_c(PyObject* iteratorable) {
	PyObject* pivot = Py_None;
	PyObject* tree[256];
	PyObject* ptr; 
	PyObject* data; 
	PyObject* xmlprolog;
	PyObject* tmp;

	char* buffer = malloc(sizeof(char)*128);
	char* pos = buffer;
	int buffer_len = 128;
	PyObject *iterator = PyObject_GetIter(iteratorable);
	int deep = 0;

	xmlprolog = PyObject_CallObject(xmlprolog_type, NULL);

	for (ptr = PyIter_Next(iterator); ptr; ptr = PyIter_Next(iterator)) {
		Py_DECREF(ptr);
		if (PyXList_Check(ptr)) {
			Satine_WriteMoreInfo("xlist %i\n", ptr->ob_refcnt);
			if (buffer != pos) {
				if (*(pos-1) == ' ')
					pos--;
				data = PyString_FromStringAndSize(buffer, pos-buffer);
				PyList_Append(pivot, data);
				Py_DECREF(data);
				pos = buffer;
			}
			Satine_WriteMoreInfo("1\n");

			if (pivot != Py_None) {
			Satine_WriteMoreInfo("1a www\n");
			Satine_WriteMoreInfo("pivot %i\n", pivot);
				PyList_Append(pivot, ptr);
			}
			else {
			Satine_WriteMoreInfo("1c\n");

				if (ptr->ob_type == (PyTypeObject*) xmldeclaration_type) {
			Satine_WriteMoreInfo("1ca\n");

					PyList_Append(xmlprolog, ptr);
					ptr = PyIter_Next(iterator);
					Py_DECREF(ptr);
					continue;
				}
				else if (ptr->ob_type == (PyTypeObject*) xmlcomment_type) {
			Satine_WriteMoreInfo("1cb\n");
					PyList_Append(xmlprolog, ptr);
					tmp = PyIter_Next(iterator);
					Py_DECREF(tmp);
					if (tmp != Py_None) {
						PyList_Append(ptr, tmp);
						Py_DECREF(PyIter_Next(iterator));
					}
					continue;
				}
			}
			Satine_WriteMoreInfo("2\n");

			tree[deep++] = ptr;
			if (deep > 255) {
				PyErr_SetString(PyExc_RuntimeError, "maximum deep in XML structure exceeded");
				return NULL;
			}
			pivot = ptr;
			
		}
		else if (ptr == Py_None) {
			if (deep == 0) 
				break;
			
			if (buffer != pos) {
				if (*(pos-1) == ' ')
					pos--;
				data = PyString_FromStringAndSize(buffer, pos-buffer);
				PyList_Append(pivot, data);
				Py_DECREF(data);
				pos = buffer;
			}

			--deep;
			if (!deep) {
				if (PyList_Size(xmlprolog))
					PyList_Insert(pivot, 0, xmlprolog);
				Py_INCREF(pivot);
				Py_DECREF(iterator);
				Py_DECREF(xmlprolog);

				free(buffer);
				return pivot;
			}
			pivot = tree[deep-1];
		}
		else {
			Satine_WriteMoreInfo("other %i\n", ptr->ob_refcnt);
			if (pivot != Py_None) 
				kernel_cache_data(ptr, &buffer, &pos, &buffer_len);
		}

	}

	free(buffer);

	PyErr_SetString(PyExc_RuntimeError, "Not well-nested XML sequence");
	return NULL;
}

PyObject*
kernel_seq2py(PyObject* self, PyObject *seq) {
	if (!PyList_Check(seq)) {
		PyErr_SetString(PyExc_TypeError, "parameter must be a list");
		return NULL;
	}
	seq = PyXList_Copy(seq);
	if (!seq)
		return NULL;
	return kernel_seq2py_c(seq);
}

PyObject*
kernel_py2seq_c(PyObject* xlist) {
	PyObject* tree[256];
	PyObject* res = PyXList_New(); 
	PyObject* iterator;
	PyObject* ptr;
	PyObject* copy;
	int deep = 0;

	for (ptr=xlist;; ptr = PyIter_Next(iterator)) {
		if (ptr == NULL) {
			PyList_Append(res, Py_None);

			--deep;
			if (!deep) 
				return res;
			else
				iterator = tree[deep];
		}
		else if (PyXList_Check(ptr)) {
			copy = PyXList_PoorCopy(ptr); 
			PyList_Append(res, copy);
			Py_DECREF(copy);

			tree[deep++] = iterator;
			iterator = PyObject_GetIter(ptr);
			if (deep > 255) {
				PyErr_SetString(PyExc_RuntimeError, "maximum deep in XML structure exceeded");
				return NULL;
			}
		}
		else {
			PyList_Append(res, ptr);
		}
		Py_DECREF(ptr);
	}

	return NULL;
}

PyObject*
kernel_py2seq(PyObject* self, PyObject *xlist) {
	if (!PyXList_Check(xlist)) {
		PyErr_SetString(PyExc_TypeError, "parameter must be a xlist");
		return NULL;
	}
	return kernel_py2seq_c(xlist);
}

PyObject*
kernel_xml2seq_c(char* s_buffer, int s_size, PyObject* suggested_uris, int options) {
	PyObject* tuple;
	PyObject* res;

	tuple = kernel_parsexml_c(s_buffer, s_size, suggested_uris, options);
	if (tuple == NULL)
		return NULL;

	if (PyTuple_GET_SIZE(tuple)) {
		res = PyTuple_GetItem(tuple,0);		
		Py_INCREF(res);
		Py_DECREF(tuple);
		Satine_WriteMoreInfo("xml2seq_c end %s %i \n", PyString_AsString(PyObject_Repr(res)), res->ob_refcnt);
		return res;
	}
	else {
		Py_DECREF(tuple);
		Py_INCREF(Py_None);
		return Py_None;
	}

}

PyObject*
kernel_xml2seq(PyObject* self, PyObject *args, PyObject* kwds) {
	PyObject* suggested_uri = NULL;
	char* s_buffer;
	int s_size;
	int options = 0;
    static char *kwlist[] = {"doc", "uris", "opt", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s#|O!i", kwlist, 
                                     &s_buffer, &s_size, &PyDict_Type, &suggested_uri,
									 &options))
		return NULL;

	Satine_WriteMoreInfo("kernel xml2seq %s\n", s_buffer);

	return kernel_xml2seq_c(s_buffer, s_size, suggested_uri, options);
}

PyObject*
kernel_xml2py(PyObject* self, PyObject *args, PyObject* kwds) {
	char* s_buffer;
	PyObject* serial_form;
	PyObject* only_items;
	PyObject* suggested_uris = NULL;
	PyObject* handlers = NULL;
	PyObject* res;
	int nsu = 0;
	int s_size;
	int options = 0;
    static char *kwlist[] = {"xml", "p2u", "opt", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s#|O!i", kwlist, 
                                     &s_buffer, &s_size, &PyDict_Type, &suggested_uris,
									 &options))
		return NULL;

	Satine_WriteMoreInfo("kernel xml2py %s\n", s_buffer);

	if (suggested_uris == NULL) {
		suggested_uris = PyDict_New();
		nsu = 1;
	}

 	serial_form = kernel_parsexml_c(s_buffer, s_size, suggested_uris, options);
	if (nsu)
		Py_DECREF(suggested_uris);
	
	if (PyTuple_GET_SIZE(serial_form)) {
		only_items = PyTuple_GetItem(serial_form, 0);
		res = kernel_seq2py_c(only_items);
		Py_DECREF(serial_form);
		return res;
	}
	else {
		Py_DECREF(serial_form);
		Py_INCREF(Py_None);
		return Py_None;
	}
}


PyObject*
kernel_py2xml(PyObject* self, PyObject *args, PyObject* kwds) {
	PyObject* suggested_prefixes = NULL;
	PyObject* declared_ns = NULL;
	PyObject* xlist;
	PyObject* res;
	int options = opt_xmlStrict;
    static char *kwlist[] = {"py", "u2p", "opt", "declared_ns", NULL};

    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O!|O!O!i", kwlist, 
                                     &PyXList_Type, &xlist, &PyDict_Type, &suggested_prefixes,
									 &PyList_Type, &declared_ns, &options))
		return NULL;

	if (!declared_ns) 
		declared_ns = PyList_New(0);
	else
		Py_INCREF(declared_ns);

	if (!suggested_prefixes) 
		suggested_prefixes = PyDict_New();
	else
		Py_INCREF(suggested_prefixes);

	res = xlist_repr_c((PyXListObject*)xlist, suggested_prefixes, declared_ns, 
			options == opt_xmlStrict);
	Py_DECREF(suggested_prefixes);
	Py_DECREF(declared_ns);

	return res;
}

PyObject*
kernel_seq2xml(PyObject* self, PyObject *args) {
	PyObject* suggested_prefixes = NULL;
	PyObject* declared_ns = PyList_New(0);
	PyObject* list;
	PyObject* res = PyString_FromString("");
	PyObject* item;
	int i, size;
	int nsp = 0;

	if (!PyArg_ParseTuple(args, "O!|O", &PyList_Type, &list, &suggested_prefixes))
		return NULL;
	if (!suggested_prefixes) {
		suggested_prefixes = PyDict_New();
		nsp = 1;
	}

	size = PyList_GET_SIZE(list);
	for (i=0; i < size; ++i) {
		item = PyList_GetItem(list, i);
		if (PyXList_Check(item))
			PyString_ConcatAndDel(&res, 
				xlist_repr_c((PyXListObject*)item, suggested_prefixes, declared_ns, 1));
		else 
			PyString_Concat(&res, item);
		
	}

	if (nsp)
		Py_DECREF(suggested_prefixes);
	Py_DECREF(declared_ns);

	return res;
}

PyObject*
kernel_strip_c(PyObject *seq) {
	int ri;
	int wi;
	int size;
	PyObject* item;
	PyObject* data;
	char* buffer = malloc(sizeof(char)*128);
	char* pos = buffer;
	int buffer_len = 128;

	size = PyList_GET_SIZE(seq);
	for (ri=0, wi=0; ri < size; ++ri) {
		item = PyList_GetItem(seq, ri);
		if (!PyString_Check(item)) {
			if (pos > buffer) {
				if (*(pos-1) == ' ')
					data = PyString_FromStringAndSize(buffer, pos-buffer-1);
				else
					data = PyString_FromStringAndSize(buffer, pos-buffer);

				PyList_SetItem(seq, wi++, data); 
				pos = buffer;
			}
			Py_INCREF(item);
			PyList_SetItem(seq, wi++, item);
		}
		else {
			kernel_cache_data(item, &buffer, &pos, &buffer_len);
		}
	}
	if (pos > buffer) {
		if (*(pos-1) == ' ')
			data = PyString_FromStringAndSize(buffer, pos-buffer-1);
		else
			data = PyString_FromStringAndSize(buffer, pos-buffer);
		PyList_SetItem(seq, wi++, data); 
		pos = buffer;
	}

	PySequence_DelSlice(seq, wi, size);
	free(buffer);
	return seq;
}

PyObject*
kernel_strip(PyObject* self, PyObject *seq) {
	if (!PyList_Check(seq)) {
		PyErr_SetString(PyExc_TypeError, "Argument must be a list");
		return NULL;
	}
	Py_INCREF(seq);
	return kernel_strip_c(seq);
}

PyObject*
kernel_dfa(PyObject* self, PyObject *seq) {
	if (!PyString_Check(seq)) {
		PyErr_SetString(PyExc_TypeError, "the parameter must be a string");
		return NULL;
	}

	return xquery_compile(seq);
}

PyObject* 
kernel_dfa_next(PyObject* self, PyObject* args) {
	PyObject* item;
	PyObject* last;
	PyObject* res;

	if (!PyArg_ParseTuple(args, "OO!", &item, &last, &PyXList_Type))
		return NULL;

	res = xquery_next(item, last);
	if (res != NULL)
		Py_INCREF(res);
	return res;
}

PyObject* 
kernel_verbose(PyObject* self, PyObject* args) {
	int verbosity = -1;
	int res;

	if (!PyArg_ParseTuple(args, "|i", &verbosity))
		return NULL;

	res = kernel_verbosity;
	if (verbosity != -1)
		kernel_verbosity = verbosity;

	return PyInt_FromLong(res);
}


static char get_type_doc[]=
"get_type(tag, uri) -> class\n\n\
Return a Python class whose instances are xlists with tag and uri.Raise an exception\n\
if a suitable class does not exist.";

static char new_type_doc[] =
"new_type(tag, module[, __attrs__[, __items__[, name]]]) -> class\n\n\
Create a new class that extends the type xlist. The class is stored in module. If tag\n\
is a reserved keyword, you have to specify the class name";

static char new_space_doc[] = 
"new_space(uri, prefix, name) -> module\n\n\
Create a new module and make it suitable to host classes with namespace uri.\n";

static char reg_space_doc[] = 
"reg_space(module, uri, prefix) -> None\n\n\
Register an existing module as a namespace with specified uri and default prefix.\n\
This function is usually not used directly, but through the function satine.xspace";

static char get_space_doc[] = 
"get_space(uri) -> dict\n\n\
Return the dictionary with all xlist types defined for the namespace uri.";

static char parsexml_doc[] =
"parsexml(doc[, suggested_prefixes]) -> (xlist, positions)\n\n\
Convert a XML string to a xlist in the serialized format, like xml2seq does. Unlike\n\
xml2seq, return a list with positions of different XML items in the original string.";

static char xml2py_doc[] =
"xml2py(doc[, suggested_prefixes]) ->xlist.\n\n\
Convert a XML string to an xlist. If the developer defined appropriate classes\n\
for the elements in the document, these classes are used to instantiate the xlist;\n\
otherwise, new classes(with no validation support) are created at run-time. The \n\
optional parameter suggested_prefixes defines the uri for each prefix in the document.\n\
If the parameter is absent and the document does not contain a xmlns attribute, the\n\
uri for each prefix is chosen according to the default prefixes. The function removes\n\
all multiple whitespace, and special characters (\\n \\t...).";

static char xml2seq_doc[] =
"xml2seq(doc[, suggested_prefixes]) -> xlist\n\n\
Converts a XML string to an xlist using serialized format. In serialized format\n\
all start tags in the XML string become empty xlists, and all end tags become\n\
None objects.";

static char py2xml_doc[] = 
"py2xml(xlist[, suggested_prefixes]) -> str\n\n\
Convert a xlist to a XML string. The parameter suggested_prefixes is a dict that defines\n\
for each uri the corresponding prefix. The parameter is useless, if the prefixes are the\n\
default ones (like in most cases)."; 

static char seq2xml_doc[] = 
"seq2xml(xlist[, suggested_prefixes]) -> str\n\n\
Convert a xlist coded in sequential format to a XML string. The parameter suggested_prefixes\n\
is a dict that defines for each uri the corresponding prefix. The parameter is useless, if \n\
the prefixes are the default ones (like in most cases)."; 

static char seq2py_doc[] = 
"seq2py(xlist) -> xlist\n\n\
Convert a xlist from the serialized format to the standard one."; 

static char py2seq_doc[] = 
"py2seq(xlist) -> xlist\n\n\
Convert a xlist from the standard format to the serialized one."; 

static char is_xlist_doc[] = 
"is_xlist(obj) -> int\n\n\
Check if obj is a xlist.";

static char strip_doc[] = 
"strip(xlist) -> xlist\n\n\
Remove double whitespaces from all strings in a xlist. It is normally used with xlists in\n\
serialized format.";

static char dfa_doc[] = 
"dfa(expression) -> list\n\n\
Create a Deterministic Finite Automate (DFA)from a XRE expression.Replacing the XRE\n\
expression with the DFA in xlist.visit could improve speed. Return the DFA root node";

static char dfa_next_doc[] = 
"xre_next(item, dfa_node) -> list\n\n\
Match item with the DFA node. If match is successful, return the next DFA node, else raise\n\
an exception.";

static char set_verbose_doc[] = 
"set_verbose(level) -> None\n\n\
Set the verbosity level. Valid values are 0, 1, 2.";

static PyMethodDef kernel_functions[] = {
	{"new_type",		kernel_new_type, 		METH_VARARGS,				new_type_doc	},
	{"get_type",		kernel_get_type, 		METH_VARARGS,				get_type_doc	},
	{"get_space",		kernel_get_space, 		METH_VARARGS,				get_space_doc	},
	{"reg_space",		kernel_reg_space, 		METH_VARARGS,				reg_space_doc	},
	{"new_space",		kernel_new_space, 		METH_VARARGS,				new_space_doc	},
	{"parsexml",		kernel_parsexml,		METH_VARARGS|METH_KEYWORDS,	parsexml_doc	}, 
	{"seq2py",			kernel_seq2py,			METH_O,						seq2py_doc		}, 
	{"py2seq",			kernel_py2seq,			METH_O,						py2seq_doc		}, 
	{"xml2seq",			kernel_xml2seq,			METH_VARARGS|METH_KEYWORDS,	xml2py_doc		}, 
	{"xml2py",			kernel_xml2py,			METH_VARARGS|METH_KEYWORDS,	xml2seq_doc		}, 
	{"py2xml",			kernel_py2xml,			METH_VARARGS|METH_KEYWORDS,	py2xml_doc		}, 
	{"seq2xml",			kernel_seq2xml,			METH_VARARGS,				seq2xml_doc		}, 
	{"is_xlist",		kernel_is_xlist, 		METH_O,						is_xlist_doc	},
	{"strip",			kernel_strip, 			METH_O,						strip_doc		},
	{"dfa",				kernel_dfa, 			METH_O,						dfa_doc			},
	{"dfa_next",		kernel_dfa_next, 		METH_O,						dfa_next_doc	},
	{"verbose",			kernel_verbose, 		METH_VARARGS,				set_verbose_doc	},
	{NULL,		NULL}		/* sentinel */
};

DL_EXPORT(void)
initkernel(void)
{
	PyObject* m;
	PyObject* d;
	PyObject* satine_uri;

	m = Py_InitModule3("kernel",
			   kernel_functions,
			   kernel__doc__);
	if (m == NULL)
		return;

	default_prefixes = PyDict_New();
	default_uris = PyDict_New();

	Py_INCREF(&PyXList_Type);
	PyModule_AddObject(m, "xlist", (PyObject *) &PyXList_Type);

	initexpatll();
	initxiter();
	initxlistinfo();
	initxlist();
	initxquery();

	srand(( unsigned ) time(NULL ));

	d = PyModule_GetDict(m);
	if (d == NULL)
		return;


	Py_INCREF(&PyXIter_Type);
	if (PyDict_SetItemString(d, "xiter",
				 (PyObject *) &PyXIter_Type) < 0)
		return;

	kernel_spaces = PyDict_New();
	Py_INCREF(kernel_spaces);
	if (PyDict_SetItemString(d, "spaces",
				 (PyObject *) kernel_spaces) < 0)
		return;


	kernel_uri2mods = PyDict_New();
	Py_INCREF(kernel_uri2mods);
	if (PyDict_SetItemString(d, "uri2mods",
				 (PyObject *) kernel_uri2mods) < 0)
		return;

	kernel_mod2uri = PyDict_New();
	Py_INCREF(kernel_mod2uri);
	if (PyDict_SetItemString(d, "mod2uri",
				 (PyObject *) kernel_mod2uri) < 0)
		return;

	Py_INCREF(default_prefixes);
	if (PyDict_SetItemString(d, "prefixes",
				 (PyObject *) default_prefixes) < 0)
		return;
		
	Py_INCREF(default_uris);
	if (PyDict_SetItemString(d, "uris",
				 (PyObject *) default_uris) < 0)
		return;


	PyModule_AddIntConstant(m, "opt_xmlComment", opt_xmlComment);
	PyModule_AddIntConstant(m, "opt_xmlEntity", opt_xmlEntity);
	PyModule_AddIntConstant(m, "opt_xmlCData", opt_xmlCData);
	PyModule_AddIntConstant(m, "opt_xmlDeclaration", opt_xmlDeclaration);
	PyModule_AddIntConstant(m, "opt_xmlPI", opt_xmlPI);
	
	
	PyModule_AddIntConstant(m, "opt_xmlStrict", opt_xmlStrict);
	PyModule_AddIntConstant(m, "opt_autoCreationOff", opt_autoCreationOff);

	satine_uri = PyString_FromString("http://satine.sourceforge.net/schemas/kernel");

	kernel_reg_space_c(m, satine_uri, 
						PyString_FromString("satine"));

	PyDict_SetItem(default_uris, PyString_FromString(""), PyString_FromString(
		"http://satine.sourceforge.net/temp/dyn/emptyuri"));
	PyDict_SetItem(default_prefixes, PyString_FromString(
		"http://satine.sourceforge.net/temp/dyn/emptyuri"), PyString_FromString(""));

	xmlcomment_type = kernel_new_type_c(satine_uri, PyString_FromString("xmlcomment"), Py_None,
		Py_None, Py_None);
	xmldeclaration_type = kernel_new_type_c(satine_uri, PyString_FromString("xmldeclaration"), Py_None,
		Py_None, Py_None);
	xmlPI_type = kernel_new_type_c(satine_uri, PyString_FromString("xmlPI"), Py_None,
		Py_None, Py_None);
	xmlprolog_type = kernel_new_type_c(satine_uri, PyString_FromString("xmlprolog"), Py_None,
		Py_None, Py_None);

}

static void
mywrite(char *name, FILE *fp, const char *format, va_list va)
{
	PyObject *file;
	PyObject *error_type, *error_value, *error_traceback;

	PyErr_Fetch(&error_type, &error_value, &error_traceback);
	file = PySys_GetObject(name);
	if (file == NULL || PyFile_AsFile(file) == fp)
		vfprintf(fp, format, va);
	else {
		char buffer[1001];
		const int written = PyOS_vsnprintf(buffer, sizeof(buffer),
						   format, va);
		if (PyFile_WriteString(buffer, file) != 0) {
			PyErr_Clear();
			fputs(buffer, fp);
		}
		if (written < 0 || written >= sizeof(buffer)) {
			const char *truncated = "... truncated";
			if (PyFile_WriteString(truncated, file) != 0) {
				PyErr_Clear();
				fputs(truncated, fp);
			}
		}
	}
	PyErr_Restore(error_type, error_value, error_traceback);
}

void
Satine_WriteInfo(const char *format, ...) {
	va_list va;

	if (kernel_verbosity) {
		va_start(va, format);
		mywrite("stdout", stdout, format, va);
		va_end(va);
	}
}

void
Satine_WriteMoreInfo(const char *format, ...) {
	va_list va;

	if (kernel_verbosity > 1) {
		va_start(va, format);
		mywrite("stdout", stdout, format, va);
		va_end(va);
	}
}

void main() {
	char* doc = "doc =\"\"\" <addressbook d:gh=\"@lambda x: x+1\" d=\"http://d\">			\n\
           <person prova=\"ni\" surname=\"Gates\" d:ame=\"Bill\">The Microsoft chief</person>\n\
           <person surname=\"Tolvald\" d:ame=\"Linus\">The Microsoft    	chief </person>\n\
         </addressbook>\"\"\"";
/*	char* doc = "doc =\"\"\" <addressbook d:gh=\"@lambda x: x+1\" d=\"http://d\">			\n\
           <person prova=\"ni\" surname=\"Gates\" d:ame=\"Bill\">The Microsoft chief</person>\n\
         </addressbook>\"\"\"";*/
	Py_Initialize();			
//	doc = "doc = '<message xmlns=\"ciao\"></> '";
	initkernel();
	PyRun_SimpleString("import sys");
	PyRun_SimpleString("import kernel");
	PyRun_SimpleString("import satine.schemas.xmlschema");
	PyRun_SimpleString(doc);
	PyRun_SimpleString("doc = open('2511B39.xsd').read()");
	PyRun_SimpleString("print kernel.xml2py(doc)");
/*	PyRun_SimpleString("for i in xrange(40000): print kernel.xml2py('<a:config d:ciao=\"hu\"/>')");
	PyRun_SimpleString("p.validate()");
	PyRun_SimpleString("gc.collect()");
	PyRun_SimpleString("i = kernel.xlist(a='1')");
	PyRun_SimpleString("j = kernel.xlist(a='2')");
	PyRun_SimpleString("print i < j");
	PyRun_SimpleString("print i.__realdict__ < j.__realdict__");
	PyRun_SimpleString("l = kernel.xlist(a='2', b='1')");
	PyRun_SimpleString("a = l.__realdict__['a']");
	PyRun_SimpleString("b = l.__realdict__['b']");
	PyRun_SimpleString("print  a < b");
//	PyRun_SimpleString("doc = '<ciao>mondo &gt; 3 </ciao>'");
	PyRun_SimpleString("d = kernel.xml2py(doc)");
	PyRun_SimpleString("print kernel.py2xml(d)");
/*	PyRun_SimpleString("print d.tpinfo.default_attrs");
	PyRun_SimpleString("print kernel.py2xml(d)");
	PyRun_SimpleString("print kernel.py2xml(d)");
//	PyRun_SimpleString("it = d.iter('..|')");
//	PyRun_SimpleString("for i in it: print 'item', i");
//	PyRun_SimpleString("for i in it.next(): print 'item', i");
	PyRun_SimpleString("print d.query('<addressbook><person>|surname')");
//	PyRun_SimpleString("for i in it: print 'item', i, it.tell()");

	PyRun_SimpleString("print d.visit([('<addressbook><person>',lambda x: 1),('<addressbook><person>',lambda x: 1)])");
	PyRun_SimpleString("import kernel");
	PyRun_SimpleString("print kernel.strip(kernel.xml2seq('<ciao>  \\n\\t  mucca    <hei>  ul  '))");
	PyRun_SimpleString(doc);*/
/*	PyRun_SimpleString("class string(kernel.xlist):\n\tdef xml2py(self,v):return v\n\tdef py2xml(self,v):return v\n");
	PyRun_SimpleString("class pattern(kernel.xlist):\n\t__attrs__='<satine:string>value'\n");
	PyRun_SimpleString("class A(kernel.xlist):\n\t__attrs__='<xsd:pattern value="\S*$">'\n");
	PyRun_SimpleString("kernel.reg_space(sys.modules['__main__'], 'dedede', 'ciao')");
	PyRun_SimpleString("a = A()");
	PyRun_SimpleString("a['ciao'] = '32'");
	PyRun_SimpleString("print a.append(kernel.xml2py('<d/>'))");
	PyRun_SimpleString("print a.validate()");
//	PyRun_SimpleString("a.pippo = '7'");
//	PyRun_SimpleString("a.prova = '7'");
//	PyRun_SimpleString("print a");
//	PyRun_SimpleString("print a.validate()");
//	PyRun_SimpleString("print a");
/*	PyRun_SimpleString("l = kernel.xml2py('<ciao><mondo/></>')");
	PyRun_SimpleString("for i in l: a1 = i");
	PyRun_SimpleString("for i in l: a2 = i");
	PyRun_SimpleString("for i in l: a3 = i");
	PyRun_SimpleString("for i in l: a4 = i");
	PyRun_SimpleString("s = kernel.xlist()");
	PyRun_SimpleString("l.append(s)");
	PyRun_SimpleString("del s[0]");

//	PyRun_SimpleString("l = kernel.xml2py()");
	PyRun_SimpleString("l.match(a)");
//	PyRun_SimpleString("print 'hei', kernel.py2xml(l)");

	PyRun_SimpleString("print kernel.xlist(ciao='2')");

	PyRun_SimpleString("prefixes = {}");
//	PyRun_SimpleString("kernel.xlist()");
//	PyRun_SimpleString("kernel.xlist().match(kernel.xlist())");
//	PyRun_SimpleString("res = kernel.xml2seq('<addressbook><person d:ame=\"Linus\">',prefixes)");
//	PyRun_SimpleString("kernel.xlist().match(res[0])");
//	PyRun_SimpleString("print res, prefixes");
//	PyRun_SimpleString("def fun(x): print 'found pattern',x");
*/
//	PyRun_SimpleString("obj = kernel.xml2py('<ciao><msg hei=\"j\"/></>')");
	PyRun_SimpleString("obj = kernel.xml2py(doc)");
//	PyRun_SimpleString("print kernel.query(obj,'<addressbook><person>?surname')");
//	PyRun_SimpleString("print kernel.query(obj,'msg',format='tag')");
	PyRun_SimpleString("for i in xrange(50): l = kernel.xlist()");
//	PyRun_SimpleString("for i in xrange(500000):  kernel.xml2seq('.*')");
//	PyRun_SimpleString("for i in xrange(100000):  kernel.dfa('.*')");
	PyRun_SimpleString("for i in xrange(36000):  obj.visit([('<addressbook>.*',lambda x: 1)])");
	PyRun_SimpleString("for k in person: print person.tell(), k");
//	PyRun_SimpleString("obj = kernel.xml2py(doc)");

//	PyRun_SimpleString("print it");

//	PyRun_SimpleString("print obj, prefixes");

//	PyRun_SimpleString("from satine import dt");
/*	PyRun_SimpleString(
		"class A(kernel.xlist):\n\
		\t__atable__={'name': dt.string}\n\
		\t__etable__=[]\n\
		\t__dtable__={'name': None}\n\
		\t__tag__='a'\n\
		\t__uri__='prova'\n\
		\tdef ciao(self): print 'ciao'\n");
	PyRun_SimpleString(
		"class A(kernel.xlist):pass");
	PyRun_SimpleString("print 'testing Satine Kernel...'");
	PyRun_SimpleString("print 'new xlist creation: a = A(21,32,8,d=2)'");
	PyRun_SimpleString("for i in range(10000):\n\ta = A(21,32,'9',d=3)\n");
	PyRun_SimpleString("for i in range(7000000):\n\ta = list([21,22])\n");
	PyRun_SimpleString("a = A(21,32,'9',d=3)");
	PyRun_SimpleString("a.append(A(21,32,'9',d=3))");
	PyRun_SimpleString("a.__dict_");
	PyRun_SimpleString("i = a.iter(tag=\"A\")");
	PyRun_SimpleString("print kernel.is_xlist(a)");
	PyRun_SimpleString("a.append(A(12, usr='9'))");
	PyRun_SimpleString("print A.__name__");
	PyRun_SimpleString("print i.next()");
	PyRun_SimpleString("i.seek(0)");
	PyRun_SimpleString("print i.next()");
	PyRun_SimpleString("for i in a: print i");
	PyRun_SimpleString("print a['d']");
	PyRun_SimpleString("ns = kernel.new_space('myuri')");
	PyRun_SimpleString("print kernel.new_type('mytag','mytag',ns)");*/
}
