/* Iterator objects */

#include <Python.h>
#include "xiterobject.h"
#include "xlistobject.h"

extern PyObject* 
kernel_xml2seq_c(char* s_buffer, int s_size, PyObject* suggested_uris);

#define DEFERRED_ADDRESS(ADDR) 0

int ctr_xquery = 0;
int ctr_function = 1;


PyObject* checkLambda(PyObject* self) {
	PyObject* cmp;
	PyObject* item;
	PyXIterObject *it = (PyXIterObject *) self;
	int res;	
		
	for (;;) {

		item = PyList_GetItem(it->it_seq, it->index++);
		if (item == NULL) {
			it->index = 0;
			PyErr_SetObject(PyExc_StopIteration, Py_None);
			return NULL;
		}

		cmp = PyObject_CallFunction((PyObject*)it->criteria, "O", item);
		if (cmp == NULL)
			return NULL;

		res = PyInt_AsLong(cmp);
		if (res)
			return item;
	}
}


PyObject* checkTag(PyObject *self) {
	PyObject* item;
	PyXIterObject *it = (PyXIterObject *) self;

	for (;;) {
		item = PyList_GetItem(it->it_seq, it->index++);
		if (item == NULL) {
			it->index = 0;
			PyErr_SetObject(PyExc_StopIteration, Py_None);
			return NULL;
		}
	
		if (PyXList_Check(item))
			if (PyObject_Compare(PyXList_GetTag(item), it->criteria) == 0)
				return item;
	}
}

PyObject* checkXRE(PyObject *self) {
	PyXIterObject *it = (PyXIterObject *) self;
	return xquery_iter(it->dfa_status);
}

PyObject *
PyXIter_New_c(PyObject *seq, PyObject* criteria, char* type) {
	PyXIterObject *it;
	it = PyObject_NEW(PyXIterObject, &PyXIter_Type);
	if (it == NULL)
		return NULL;

	it->index = 0;
	it->criteria = criteria;
	it->dfa_status = NULL;
	
	Py_INCREF(seq);
	it->it_seq = seq;

	if (!criteria)
		return (PyObject *)it;

	Py_INCREF(criteria);
	if (!strcmp(type, "xre")) {
		it->dfa_status = xquery_init_iter(seq, criteria);
		if (it->dfa_status == NULL)
			return NULL;
		it->checkFunction = checkXRE;
	}
	else if (!strcmp(type, "pyfun"))
		it->checkFunction = checkLambda;
	else if (!strcmp(type, "tag"))
		it->checkFunction = checkTag;
	else {
		PyErr_SetString(PyExc_ValueError, "format is not supported");
		return NULL;
	}

	return (PyObject *)it;
}

static void
xiter_dealloc(PyXIterObject *it)
{
	PyObject_GC_Fini(it);
	Py_DECREF(it->it_seq);
	Py_XDECREF(it->criteria);

	if (it->dfa_status) 
		xquery_free_iter(it->dfa_status);
		
	it = (PyXIterObject *) PyObject_AS_GC(it);
	PyObject_DEL(it);
}

static int
xiter_traverse(PyXIterObject *it, visitproc visit, void *arg)
{
	return visit(it->it_seq, arg);
}

static PyObject *
xiter_simple_next(PyXIterObject *it)
{
	PyObject *seq = (PyObject*) it->it_seq;
	PyObject *result = PySequence_GetItem(seq, it->index++);

	if (result == NULL && PyErr_ExceptionMatches(PyExc_IndexError))
		PyErr_SetObject(PyExc_StopIteration, Py_None);
	return result;
}


static PyObject *
xiter_complex_next(PyXIterObject *it)
{
	PyObject* res;

	res = it->checkFunction((PyObject*)it);
	
	if (res == NULL && PyErr_ExceptionMatches(PyExc_StopIteration)) 
		it->index = 0;
	
	return res;
}

static PyObject *
xiter_next(PyXIterObject *it)
{
	if (!it->criteria)
		return xiter_simple_next(it);
	else
		return xiter_complex_next(it);
}

static PyObject *
xiter_getiter(PyObject *it)
{
	Py_INCREF(it);
	return it;
}

static PyObject *
xiter_iternext(PyObject *iterator)
{
	return xiter_next((PyXIterObject*) iterator);
}

static PyObject *
xiter_tell(PyObject *iterator) {
	PyXIterObject *it;
	PyObject *seq;
	int index;

	assert(PyXIter_Check(iterator));
	it = (PyXIterObject *)iterator;
	seq = it->it_seq;

	if (it->checkFunction == checkXRE)
		index = xquery_iter_tell(it->dfa_status, 1);
	else
		index = it->index;
	return PyInt_FromLong(index);
}

static PyObject *
xiter_seek(PyObject *iterator, PyObject* index) {
	PyXIterObject *it;
	int idx;

	assert(PyXIter_Check(iterator));
	it = (PyXIterObject *)iterator;

	if (PyInt_Check(index)) {
		idx = PyInt_AsLong(index);

		if (it->checkFunction == checkXRE)
			xquery_iter_seek(it->dfa_status, 1, idx, 1);
		else
			it->index = idx;

		Py_INCREF(Py_None);
		return Py_None;
	}
	else {
		PyErr_SetString(PyExc_TypeError, "index must be an integer");
		return NULL;
	}

}

static char next_doc[] = 
"it.next()->obj\n\n\
Get the next value, or raise StopIteration";

static char tell_doc[] = 
"it.tell()->int\n\n\
Return the iterator index in the original xlist. The index deals only with direct subitems\n\
of the original xlist";

static char seek_doc[] = 
"it.seek(pos)->int\n\n\
Set the iterator index at pos. Following objects are extracted starting from pos in the\n\
original xlist";


static PyMethodDef xiter_methods[] = {
	{"next",	(PyCFunction)xiter_next,	METH_NOARGS, next_doc	},
	{"tell",	(PyCFunction)xiter_tell,	METH_NOARGS, tell_doc	},
	{"seek",	(PyCFunction)xiter_seek,	METH_O,		 seek_doc	},
	{NULL,		NULL}		/* sentinel */
};

static char xiter_doc[] =
"please use the method xiter of a xlist or the operatore iter to get the iterator";

PyTypeObject PyXIter_Type = {
	PyObject_HEAD_INIT(DEFERRED_ADDRESS(&PyType_Type))
	0,					/* ob_size */
	"xiterator",				/* tp_name */
	sizeof(PyXIterObject) + PyGC_HEAD_SIZE,	/* tp_basicsize */
	0,					/* tp_itemsize */
	/* methods */
	(destructor)xiter_dealloc, 		/* tp_dealloc */
	0,					/* tp_print */
	0,					/* tp_getattr */
	0,					/* tp_setattr */
	0,					/* tp_compare */
	0,					/* tp_repr */
	0,					/* tp_as_number */
	0,					/* tp_as_sequence */
	0,					/* tp_as_mapping */
	0,					/* tp_hash */
	0,					/* tp_call */
	0,					/* tp_str */
	PyObject_GenericGetAttr,		/* tp_getattro */
	0,					/* tp_setattro */
	0,					/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_GC,	/* tp_flags */
 	xiter_doc,					/* tp_doc */
 	(traverseproc)xiter_traverse,		/* tp_traverse */
 	0,					/* tp_clear */
	0,					/* tp_richcompare */
	0,					/* tp_weaklistoffset */
	(getiterfunc)xiter_getiter,		/* tp_iter */
	(iternextfunc)xiter_iternext,		/* tp_iternext */
	xiter_methods,				/* tp_methods */
	0,					/* tp_members */
	0,					/* tp_getset */
	0,					/* tp_base */
	0,					/* tp_dict */
	0,					/* tp_descr_get */
	0,					/* tp_descr_set */
};

void initxiter() {
	PyXIter_Type.ob_type = &PyType_Type;
	if (PyType_Ready(&PyXIter_Type) < 0)
		return;
}

