#include <Python.h>
#include "kernel.h"
#include "xlistobject.h"
#include "xiterobject.h"


static PyObject* pt_text;
static PyObject* pt_any;

PyObject* xquery_next(PyObject* item, PyObject* last) {
	int i, size;
	PyObject* edge;
	PyObject* edge_label;
	PyObject* edge_next;
	PyObject* any = NULL;

	size = PyList_GET_SIZE(last);
	for (i=0; i < size; ++i) {
		edge = PyList_GET_ITEM(last, i);
		edge_label = PyTuple_GET_ITEM(edge, 0);
		edge_next = PyTuple_GET_ITEM(edge, 1);
		
		if (edge_label == pt_text) {
			if (!PyXList_Check(item))
				return edge_next;
		}
		else if (edge_label == pt_any) {
			any = edge_next;
		}		
		else if (PyXList_Check(item)) {
			if (xlist_match_c((PyXListObject*) item,(PyXListObject*) edge_label)) 
				return edge_next;
		}
	}
	return any;
}


static 
PyObject* xquery_new_node(PyObject* nodes) {
//	PyObject* node = PyObject_CallObject((PyObject*) &PyXList_Type, NULL);
	PyObject* node = PyXList_New();
	PyList_Append(nodes, node);
//	Py_DECREF(node);
	PyObject_SetAttrString(node, "final", Py_False);
	return node;
}

static 
void xquery_del_edge(PyObject* edge) {
	Py_DECREF(edge);
}

static
void xquery_link(PyObject* source, PyObject* dest, PyObject* label) {
	int size = PyList_GET_SIZE(source);
	PyObject* edge;

	edge = PyTuple_New(2);

	Py_INCREF(label);
	
	PyTuple_SET_ITEM(edge, 0, label);
	PyTuple_SET_ITEM(edge, 1, dest);
	Py_INCREF(dest);
	PyList_Append(source, edge);
	Py_DECREF(edge);
}

static
PyObject* xquery_get_edge(PyObject* source, PyObject* label) {
	PyObject* edge;
	PyObject* edge_label;
	PyObject* next_node;
	int i;
	int size = PyList_GET_SIZE(source);

	for (i=0; i < size; ++i) {
		edge = PyList_GET_ITEM(source, i);
		edge_label = PyTuple_GET_ITEM(edge, 0);
		next_node = PyTuple_GET_ITEM(edge, 1);

		if (edge_label == label) {
			return next_node;
		}
	}
	return NULL;
}

static
PyObject* xquery_add_node(PyObject* source, PyObject* label, PyObject* nodes) {
	PyObject* edge;
	PyObject* next_node;

	next_node = xquery_new_node(nodes);

	edge = PyTuple_New(2);

	Py_INCREF(label);

	PyTuple_SET_ITEM(edge, 0, label);
	PyTuple_SET_ITEM(edge, 1, next_node);
	PyList_Append(source, edge);
	Py_DECREF(edge);
	return next_node;
}

static
PyObject* xquery_seq(int* idx, char* model, PyObject** patterns, PyObject* root, PyObject* nodes);

static
PyObject* xquery_term(int* idx, char* model, PyObject** patterns, PyObject* root, PyObject* nodes) {
	PyObject* seq;
	char _$ = model[*idx];
	char err[2];
	(*idx)++;

	switch (_$) {
		case  '(': {
			seq = xquery_seq(idx, model, patterns, root, nodes);
			if (model[*idx] != ')')
				return NULL;
			else {
				(*idx)++;
				return seq;
			}
				   }
		case  'x': {
			return xquery_add_node(root, patterns[*idx-1], nodes); 
				   }
		case  '.': {
			return xquery_add_node(root, pt_any, nodes); 
				   }
		case  '$': {
			return xquery_add_node(root, pt_text, nodes); 
				   }
		default: 
			err[0] = _$;
			err[1] = 0;
			Satine_WriteInfo("invalid character '%s' in XRE %s", err, model);
			PyErr_SetObject(PyExc_AttributeError, 
				PyString_FromFormat("invalid character '%s' in XRE %s", err, model));
			return NULL;
	}
	return root;
}

static
PyObject* xquery_op(int* idx, char* model, PyObject** patterns, PyObject* root, PyObject* nodes) {
	char _$;
	
	PyObject* _$1 = xquery_term(idx, model, patterns, root, nodes);
	if (_$1 == NULL)
		return NULL;

	_$ = model[*idx];
	switch (_$) {
		case '*': {
			(*idx)++;
			xquery_link(_$1, root, Py_None);
			xquery_link(root, _$1, Py_None);
			return _$1;
		}
		case '?': {
			(*idx)++;
			xquery_link(root, _$1, Py_None);
			return _$1;
		}
		case ('+') : {
			(*idx)++;
			xquery_link(_$1, root, Py_None);
			xquery_link(root, _$1, Py_None);
			return _$1;
		}
		default: return _$1;
	}
}

static
PyObject* xquery_seq(int* idx, char* model, PyObject** patterns, PyObject* root, PyObject* nodes) {
	PyObject* next;
	PyObject* last;
	PyObject* second_last;
	PyObject* third_last;
//	PyObject* _$1;
	PyObject* _$2;
//	PyObject* _R1;
	PyObject* _R2;
	char _$;

	third_last = NULL;
	second_last = NULL;
	last = root;
	for (;;) {
		_$ = model[*idx];
		if (_$ == 0) {
			PyObject_SetAttrString(last, "final", Py_True);
			return last;
		}
		else if  (_$ == ')') {
			return last;
		}
		else if (_$ == '|') {
			if (third_last == NULL) {
				Satine_WriteInfo("pipe without previous expression in XRE %s at position %i", 
					model, idx);
				PyErr_SetObject(PyExc_AttributeError, 
					PyString_FromFormat("pipe without previous expression in XRE %s at position %i", 
					model, idx));
				return NULL;
			}

			(*idx)++;
			_R2 = xquery_add_node(third_last, Py_None, nodes);
			_$2 = xquery_op(idx, model, patterns, _R2, nodes);
			if (!_$2)
				return NULL;
			next = xquery_new_node(nodes);
			xquery_link(last, next, Py_None);
			xquery_link(_$2, next, Py_None);
			last = next;
		}
		else {
			third_last = last;
			last = xquery_add_node(last, Py_None, nodes);
			second_last = last;
			last = xquery_op(idx, model, patterns, last, nodes);	
			if (!last)
				return NULL;

			_$ = model[*idx];
		}
	}
}

static
int xquery_edgeInside(PyObject* node, PyObject* _edge) {
	int size, i;
	PyObject* edge;
	PyObject* edge_label;
	PyObject* _edge_label;

	size = PyList_GET_SIZE(node);
	_edge_label = PyTuple_GetItem(_edge, 0);
	for (i=0; i < size; ++i) {
		edge = PyList_GET_ITEM(node, i);
		edge_label = PyTuple_GET_ITEM(edge, 0);

		if (_edge_label == edge_label) 
			return 1;
	}

	return 0;

}

static
void xquery_remove_epsilon(PyObject* node) {
	int size, i, j, nsize;
	PyObject* edge;
	PyObject* edge_label;
	PyObject* next_node;
	PyObject* final;
	PyObject* next_node_edge;

//		PySys_WriteStdout("%s\n\n", PyString_AsString(PyObject_Str(node)));
	size = PyList_GET_SIZE(node);
	for (i = 0; i < size; ++i) {
		edge = PyList_GET_ITEM(node, i);
		edge_label = PyTuple_GET_ITEM(edge, 0);
		next_node = PyTuple_GET_ITEM(edge, 1);
		
		if (edge_label == Py_None) {
			PySequence_DelItem(node, i--);
			xquery_remove_epsilon(next_node);
	//		PySys_WriteStdout("%s\n\n", PyString_AsString(PyObject_Str(next_node)));

			final = PyObject_GetAttrString(next_node, "final");
			if (PyInt_AsLong(final))
				PyObject_SetAttrString(node, "final", final);
			Py_DECREF(final);

			nsize = PyList_GET_SIZE(next_node);
			for (j = 0; j < nsize; ++j) {
				next_node_edge = PyList_GET_ITEM(next_node, j);
				if (!xquery_edgeInside(node, next_node_edge))
					PyList_Append(node, next_node_edge);
			}
//			PySys_WriteStdout("%s\n\n", PyString_AsString(PyObject_Str(node)));


			size--;
		}

//			PySys_WriteStdout("%s\n\n", PyString_AsString(PyObject_Str(node)));		
	}

}

static
void xquery_epsilon(PyObject* nodes) {
	int j;
	int nsize;
	PyObject* node;

	nsize = PyList_GET_SIZE(nodes);
/*	for (j = 0; j < nsize; ++j) {
		node = PyList_GET_ITEM(nodes, j);
		Py_INCREF(node);
	}	*/
	for (j = 0; j < nsize; ++j) {
		node = PyList_GET_ITEM(nodes, j);
		xquery_remove_epsilon(node);
	}
/*	for (j = 0; j < nsize; ++j) {
		node = PyList_GET_ITEM(nodes, j);
		Py_DECREF(node);
	}*/
}

static
PyObject* xquery_graph(PyObject* rule) {
	int i, j, len, index;
	PyObject* item;
	PyObject* root;
	PyObject* res;
	PyObject* nodes;

	char model[256];
	PyObject* patterns[256];
	char* tmp;


	item = PyList_GetItem(rule, 0);
	for (i=1, j=0; item; ++i) {
		if (j >= 256) {
			return NULL;
		}

		if (PyXList_Check(item)) { 
			patterns[j] = item;
			model[j++] = 'x';
		}
		else if (PyString_Check(item)) {
			PyString_AsStringAndSize(item, &tmp, &len);
			if (j+len > 256) {
				return NULL;
			}
			strncpy(model+j, tmp, len);
			j += len;
		}
		item = PyList_GetItem(rule, i);
	}
	model[j] = 0;


	index = 0;
	nodes = PyList_New(0);
	root = xquery_new_node(nodes);
	res = xquery_seq(&index, model, patterns, root, nodes);
//	PySys_WriteStdout("root is %s\n\n", PyString_AsString(PyObject_Repr(root)));
	xquery_epsilon(nodes);
//	item = PyList_GetItem(root, 1);
//	for (i=0; i < PyList_GET_SIZE(root); ++i) {
//		item = PyList_GetItem(root, i);
//		PySys_WriteStdout("item %s ref %i\n\n", PyString_AsString(PyObject_Repr(item)), item->ob_refcnt);
//	}
//	PySys_WriteStdout("root is %s\n\n", PyString_AsString(PyObject_Repr(root)));

	Py_DECREF(nodes);

//	PySys_WriteStdout("nodes are %s\n\n", PyString_AsString(PyObject_Repr(nodes)));
/*	for (i=0; i < PyList_GET_SIZE(nodes); ++i) {
		item = PyList_GetItem(nodes, i);
//		PySys_WriteStdout("item %s ref %i\n\n", PyString_AsString(PyObject_Repr(item)), item->ob_refcnt);
	}
*/	
	if (res == NULL)
		return NULL;
	else 
		return root;
}


PyObject* xquery_compile(PyObject* expression) {
	char* expression_c;
	int expression_l;
	PyObject* compiled;
	PyObject* res;

	if (PyString_Check(expression)) {
		PyString_AsStringAndSize(expression, &expression_c, &expression_l);
		compiled = kernel_xml2seq_c(expression_c, expression_l, NULL, 0);
		if (compiled == NULL) {
			return NULL;
		}
		compiled = kernel_strip_c(compiled);
		if (compiled == NULL) {
			return NULL;
		}
	}
	else if (PyList_Check(expression)) {
		Py_INCREF(expression);
		compiled = expression;
	}
	else {
		return NULL;
	}

	res = xquery_graph(compiled);
	Py_DECREF(compiled);
	return res;
}


QueryInfo*
xquery_compile_q(PyObject* query) {
	int i, j;
	char* query_c;
	char c;
	int query_s;
	QueryInfo* info;
	PyObject* compiled;
	
	if (PyString_AsStringAndSize(query, &query_c, &query_s) == -1)
		return NULL;

	info = (QueryInfo*) malloc(sizeof(QueryInfo));
	info->params = PyList_New(0);
	for (j = i = query_s-1; i >= 0; --i) {
		c = query_c[i];
		if (c == ',' || c == '>' || c == '|') {
			if (i < j) 
				PyList_Append(info->params, PyString_FromStringAndSize(query_c+i+1, j-i));
			j = i;
				
			if (c != ',')
				break;
		}
	}

	compiled = kernel_xml2seq_c(query_c, i, NULL, 0);

	if (!compiled) {
		Py_DECREF(info->params);
		free(info);
		PyErr_SetString(PyExc_ValueError, "The query pattern is not a valid XML fragment");
		return NULL;
	}

	if (PyList_GET_SIZE(compiled) == 0) {
		Py_DECREF(compiled);
		Py_DECREF(info->params);
		free(info);
		PyErr_SetString(PyExc_ValueError, "The query pattern is not valid");
		return NULL;
	}

	info->pattern = xquery_graph(compiled);
	if (!info->pattern) {
		Py_DECREF(compiled);
		Py_DECREF(info->params);
		free(info);
		return NULL;
	}

	Py_DECREF(compiled);
	return info;
}


QueryStatus*
xquery_init_iter(PyObject* xlist, PyObject* query_expr) {
	QueryInfo* query;
	QueryStatus* status;
	query = xquery_compile_q(query_expr);
	if (query == NULL)
		return NULL;

	status = malloc(sizeof(QueryStatus));
	status->xlist = xlist;
	status->query_nodes[0] = query->pattern;
	status->deep = 0;
	status->params = query->params;
	free(query);
	return status;
}


void
xquery_free_iter(QueryStatus* status) {
	Py_DECREF(status->params);
	Py_DECREF(status->query_nodes[0]);
	for (; status->deep; status->deep--)
		Py_DECREF(status->iterators[status->deep]);

	free(status);
}

int 
xquery_iter_tell(QueryStatus* status, int deep) {
	PyXIterObject * iter;
		
	if (deep == 0 || deep > status->deep)
		return -1;

	iter = (PyXIterObject *) status->iterators[deep];
	return iter->index;
}

void
xquery_iter_seek(QueryStatus* status, int deep, int index, int clear) {
	PyXIterObject * iter = (PyXIterObject *) status->iterators[deep];
	iter->index = index;

	if (clear) {
		for (;status->deep > deep; status->deep--) 
			Py_DECREF(status->iterators[status->deep]);
	}
}

PyObject*
xquery_next_match(QueryStatus* status) {
	PyObject* ptr; 
	PyObject *iterator;
	PyObject* last_match;
	PyObject* final;
	PyObject* match;
		
	last_match = status->query_nodes[status->deep];
	if (status->deep == 0) 
		ptr = status->xlist;
	else {
		iterator = status->iterators[status->deep];
		ptr = PyIter_Next(iterator);
	}
	
	for (;;ptr = PyIter_Next(iterator)) {
		if (ptr == NULL) {
			--status->deep;
			Py_DECREF(iterator);
			iterator = status->iterators[status->deep];
			last_match = status->query_nodes[status->deep];
		}
		else {
			match = xquery_next(ptr, last_match);
			if (match) {
				last_match = match;
				if (PyXList_Check(ptr)) {
					iterator = PyObject_GetIter(ptr);
					status->deep++;
					status->iterators[status->deep] = iterator;
					status->query_nodes[status->deep] = match;
					
					if (status->deep > 255) {
						PyErr_SetString(PyExc_RuntimeError, "maximum deep in XML structure exceeded");
						return NULL;
					}
				}

				final = PyObject_GetAttrString(match, "final");
				if (PyInt_AsLong(final)) {
					Py_DECREF(final);
					return ptr;
				}
				Py_DECREF(final);
			}
		}
		if (status->deep == 0)
			break;
	}
	
	PyErr_SetObject(PyExc_StopIteration, Py_None);
	return NULL;
}

PyObject* xquery_iter(QueryStatus* status) {
	PyObject* res = Py_None;
	PyObject* params;
	int params_len;
	int j;
	PyObject* key;
	PyObject* value;
	PyObject* match;

	match = xquery_next_match(status);

	if (match == NULL) {
		return NULL;
	}

	if (!PyXList_Check(match)) {
		Py_INCREF(match);
		return match;
	}

	params = status->params;
	params_len = PyList_GET_SIZE(params);
	if (params_len == 1) {
		key = PyList_GetItem(params, 0);
		res = PyDict_GetItem(((PyXListObject*)match)->attrs, key);
		if (res) {
			Py_INCREF(res);
		}
		else {
			Py_INCREF(Py_None);
			res = Py_None;
		}
	}
	else if (params_len > 1) {
		res = PyTuple_New(params_len);
		for (j=0; j < params_len; ++j) {
			key = PyList_GetItem(params, j);
			value = PyDict_GetItem(((PyXListObject*)match)->attrs, key);
			if (value) {
				PyTuple_SetItem(res, j, value);
			} else {
				Py_INCREF(Py_None);
				PyTuple_SetItem(res, j, Py_None);
			}
		}
	}
	else {
		Py_INCREF(match);
		res = match;
	}

	return res;
}

static int
xquery_init_visit(PyObject* items, PyObject** queryInfos, PyObject* handlers) {
	int pos;
	PyObject* item;
	PyObject* pattern;
	PyObject* queryInfo;
	PyObject* method;

	int size = PyList_GET_SIZE(items);

	for (pos = 0; pos < size; ++pos) {
		item = PyList_GET_ITEM(items, pos);
		pattern = PyTuple_GET_ITEM(item, 0);
		if (!PyList_Check(pattern))
			queryInfo = xquery_compile(pattern);
		else {
			queryInfo = pattern;
			Py_INCREF(queryInfo);
		}
		
		if (!queryInfo) 
			return 0;

		method = PyTuple_GetItem(item, 1);

		if (!method || !PyCallable_Check(method)){
			PyErr_SetString(PyExc_AttributeError, "bad method");
			return 0;
		}

//		Py_INCREF(queryInfo);
		queryInfos[pos] = queryInfo;
		Py_INCREF(method);
		PyTuple_SET_ITEM(handlers, pos, method);

	}
	return 1;
}



PyObject* 
xquery_visit(PyObject* xlist, PyObject* items) {
	int size = PyList_GET_SIZE(items);
	PyObject** queryInfos[256];
	PyObject* handlers;
	PyObject* iterators[256];
	PyObject* iterator = NULL;
	PyObject* ptr;
	PyObject* match;
	PyObject* last_match;
	PyObject* final;
	int deep = 0;
	int deeper = 0;
	int i;

	handlers = PyTuple_New(size);

	for (i=0; i < 256; ++i)
		queryInfos[i] = NULL;

	queryInfos[0] = malloc(sizeof(PyObject*)*size);
	if (!xquery_init_visit(items, queryInfos[0], handlers)) {
		free(queryInfos[0]);
		Py_DECREF(handlers);
		return NULL;
	}
		
	for (ptr = xlist;;ptr = PyIter_Next(iterator)) {
		if (ptr == NULL) {
			if (deep) {
				deep--;
				iterator = iterators[deep];
			}
			else 
				break;
		}
		else {
			for (i=0; i < size; ++i) {
				last_match = queryInfos[deep][i];
				if (last_match == NULL)
					continue;

				if (queryInfos[deep+1] == NULL)
					queryInfos[deep+1] = malloc(sizeof(PyObject*)*size);

				match = xquery_next(ptr, last_match);

				if (match) {
					queryInfos[deep+1][i] = match;
					final = PyObject_GetAttrString(match, "final");
					if (PyInt_AsLong(final)) {
						PyObject_CallFunction(PyTuple_GET_ITEM(handlers, i), "O", ptr);
					}
					else if (PyXList_Check(ptr)) {
						deeper = 1;
					}
					Py_DECREF(final);
				}
			}

			if (deeper) {
				deeper = 0;
				iterator = PyObject_GetIter(ptr);
				iterators[deep] = iterator;
				deep++;
				
				if (deep > 255) {
					PyErr_SetString(PyExc_RuntimeError, "maximum deep in XML structure exceeded");
					return NULL;
				}
			}
			if (iterator == NULL)
				break;
			
		}
	}

	for (i=0; i < size; ++i) {
		Py_DECREF(queryInfos[0][i]);
	}

	for (i=0; i < 256; ++i) 
		if (queryInfos[i] != NULL)
			free(queryInfos[i]);
		else 
			break;

	if (iterator != NULL)
		Py_DECREF(iterator);

	Py_DECREF(handlers);

	
	Py_INCREF(Py_None);
	return Py_None;
}


/*
t: x
   ( seq )

op: t+
    t* 
    t? 
	t


seq: op | op seq
	 op seq
	 $

*/


void initxquery() {
	pt_text = PyString_FromString("#text");
	pt_any = PyString_FromString("#any");

}